/*
	TaltOS
	on the Philips Chip LPC2106

	By Peter Eordegh
*/

#include "lpc2106_memorymap.h"
#include "lpc2106_functions.h"
#include "TaltOS_defs.h"
#include "TaltOS.h"


void TaltOS_clear_tasklist()
{
	TaltOS_task* temp;
	unsigned int i;
	for(i = 0; i < TALTOS_TASK_LIST_SIZE; i++)
	{
		temp = ((TaltOS_task*)(TALTOS_TASK_LIST_PTR)) + i;
		temp->address  = TALTOS_TASK_NOADDRESS;
		temp->priority = TALTOS_TASK_NULL_PRIORITY;
		temp->state    = TALTOS_TASK_OFF;
		temp->argument = TALTOS_TASK_NULL_ARGUMENT;
	}
}

//Clear the flash operation buffer
void TaltOS_clear_flash_op_buffer()
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	unsigned int i;
	for(i = 0; i < (TALTOS_FLASH_OP_BUFFER_SIZE / sizeof(unsigned long)); i++)
	{
		(*((unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR + i)) = TALTOS_EMPTY;
	}
	env->flash_op_buffer_ptr = 0;
}


//Clear the input buffer
void TaltOS_clear_input_buffer()
{
	unsigned int i;
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	for(i = 0; i < (TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long)); i++)
	{
		(*((unsigned long*)TALTOS_INPUT_BUFFER_PTR + i)) = TALTOS_EMPTY;
	}
	env->input_buf_start = 0;
	env->input_buf_end = 0;
	env->input_buf_act_size = 0;
}

//Clear the output buffer
void TaltOS_clear_output_buffer()
{
	unsigned int i;
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	for(i = 0; i < (TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long)); i++)
	{
		(*((unsigned long*)TALTOS_INPUT_BUFFER_PTR + i)) = TALTOS_EMPTY;
	}
	env->output_buf_start = 0;
	env->output_buf_end = 0;
	env->output_buf_act_size = 0;
}

//Clear the environement variables of the TalOS
void TaltOS_clear_environement()
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	env->input_buf_start	= 0;
	env->input_buf_end		= 0;
	env->input_buf_act_size = 0;
	env->output_buf_start	= 0;
	env->output_buf_end		= 0;
	env->output_buf_act_size= 0;
	
	env->actual_task = TALTOS_TASK_NULL_PTR;
	
	env->TaltOS_starting_address = TaltOS_get_component_address(TALTOS_ID1, TALTOS_ID2);
	
	env->component_receiving_state = COMPONENT_NOT_RECEIVING;
	env->component_receiving_size = 0;
	
	env->removing_component_ID1 = 0;
	env->removing_component_ID2 = 0;
	
}

//The TaltOS initialization routine
void TaltOS_init()
{
	DEBUG('T')
	TaltOS_clear_flash_op_buffer();
	TaltOS_clear_input_buffer();
	TaltOS_clear_output_buffer();
	TaltOS_clear_environement();
	TaltOS_clear_tasklist();
	//LPC2106 specific code
	REG8(TIMER0_TCR) = 0x1;
	lpc2106_vectored_interrupts_setup();
	//End of LPC2106 specific code
	TaltOS_run_and_list_components(TALTOS_RUN_AND_LIST_SHOULDLIST | TALTOS_RUN_AND_LIST_SHOULDRUN);
	TaltOS_loop();
}

//The TaltOS main loop
void TaltOS_loop()
{
	while(1)
	{
		TaltOS_set_run_end_next_task();
	}
}

//Searches, runs and ends the next task in the task list (concerning prioroty)
void TaltOS_set_run_end_next_task()
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	TaltOS_task* temp_task;
	unsigned int i;
	
	
	for(i = 0; i < TALTOS_TASK_LIST_SIZE; i++)
	{
		temp_task = (TaltOS_task*)TALTOS_TASK_LIST_PTR + i;
		if (temp_task->state == TALTOS_TASK_READYTORUN)
		{
			if (env->actual_task == TALTOS_TASK_NULL_PTR)
			{
				env->actual_task = temp_task;
			}
			else
			{
				if (env->actual_task->priority > temp_task->priority)
				{
					env->actual_task = temp_task;
				}
			}
		}
	}
	
	if (!env->actual_task == TALTOS_TASK_NULL_PTR)
	{
		void (*task)(TaltOS_task_parameters) = (void(*)(TaltOS_task_parameters))env->actual_task->address;
		TaltOS_task_parameters to_pass;
		
		// Set the parameters of the task
		to_pass.TaltOS_get_component_address_ptr = TaltOS_get_component_function(env->TaltOS_starting_address, TALTOS_GET_COMPONENT_ADDRESS);
		to_pass.TaltOS_get_component_function_ptr = TaltOS_get_component_function(env->TaltOS_starting_address, TALTOS_GET_COMPONENT_FUNCTION);
		to_pass.argument = env->actual_task->argument;
		
		// Set the task's state to running
		env->actual_task->state = TALTOS_TASK_RUNNING;
		// Run the actual task
		task(to_pass);
		// Ending of the actual task
		env->actual_task->state    = TALTOS_TASK_OFF;
		env->actual_task->address  = TALTOS_TASK_NOADDRESS;
		env->actual_task->priority = TALTOS_TASK_NULL_PRIORITY;
		env->actual_task->argument = TALTOS_TASK_NULL_ARGUMENT;
		env->actual_task = TALTOS_TASK_NULL_PTR;
	}
}

//The Task adding function
unsigned int TaltOS_add_task(unsigned long address, unsigned short priority, unsigned long argument)
{
	TaltOS_task* temp_task;
	unsigned int i;
	for(i = 0; i < TALTOS_TASK_LIST_SIZE; i++)
	{
		temp_task = (TaltOS_task*)TALTOS_TASK_LIST_PTR + i;
		if (temp_task->state == TALTOS_TASK_OFF)
		{
			temp_task->address  = address;
			temp_task->priority = priority;
			temp_task->argument = argument;
			temp_task->state    = TALTOS_TASK_READYTORUN;
			return TALTOS_TRUE;
		}
	}
	return TALTOS_FALSE;
}

void TaltOS_run_and_list_components_task(TaltOS_task_parameters parameters)
{
	TaltOS_run_and_list_components(parameters.argument);
}

//Run and list all the components' init function (needed after a reset)
void TaltOS_run_and_list_components(unsigned long should_do)
{
	TaltOS_reg_entry *comp_reg;
	unsigned int i;
	unsigned int should_run = 0;
	unsigned int should_list = 0;
	
	if (should_do & TALTOS_RUN_AND_LIST_SHOULDRUN)
	{
		should_run = 1;
	}
	if (should_do & TALTOS_RUN_AND_LIST_SHOULDLIST)
	{
		should_list = 1;
	}
	
	for(i = 0; i < TALTOS_COMPONENT_TABLE_SIZE / sizeof(TaltOS_reg_entry); i++)
	{
		comp_reg = (TaltOS_reg_entry*)TALTOS_COMPONENT_TABLE_PTR + i;
		if((comp_reg->ID1 != TALTOS_EMPTY) && (comp_reg->ID2 != TALTOS_EMPTY))
		{
			if (should_list)
			{
				TaltOS_long_output_write(TALTOS_CRLFCRLF);
				TaltOS_long_output_write(comp_reg->ID1 + TALTOS_00000000);
				TaltOS_long_output_write(comp_reg->ID2 + TALTOS_00000000);
			}
			if (should_run)
			{
				if(	!((comp_reg->ID1 == TALTOS_ERASED) && (comp_reg->ID2 == TALTOS_ERASED)) &&
					!((comp_reg->ID1 == TALTOS_ID1) && (comp_reg->ID2 == TALTOS_ID2)) &&
					!(TaltOS_is_garbage_reg(comp_reg)))
				{
					unsigned long init_function_address = TaltOS_get_component_function(comp_reg->address, COMP_INIT);
					if (!TaltOS_add_task(init_function_address, TALTOS_INIT_TASK_PRIORITY, TALTOS_TASK_NULL_ARGUMENT)) TaltOS_long_output_write(TALTOS_TF_CRLF);
					//else TaltOS_long_output_write(TALTOS_TA_CRLF);
				}
			}
		}
	}
	TaltOS_long_output_write(TALTOS_OK_CRLF);
}


//The TaltOS tic routine for the preemtive task change
void TaltOS_tic()
{
	//REG8(UART0_THR) = 'p';
}

//The input buffer's input function, can be called by the microcontroller
void TaltOS_long_input_write(unsigned long inlong)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	unsigned long* address = (unsigned long*)(TALTOS_INPUT_BUFFER_PTR) + env->input_buf_end;
	unsigned long  comp_address;
	(*address) = inlong;

	if (env->input_buf_end < (TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long)))
	{
		if (env->input_buf_start != env->input_buf_end + 1){env->input_buf_end++;}
	}
	else
	{
		if (env->input_buf_start != 0){env->input_buf_end = 0;}
	}

	(env->input_buf_act_size)++;

	comp_address = TaltOS_get_component_function(env->TaltOS_starting_address, TALTOS_CHECK_ORDER);
	
	if (!TaltOS_add_task(comp_address, TALTOS_OWN_TASK_PRIORITY, TALTOS_TASK_NULL_ARGUMENT)) TaltOS_long_output_write(TALTOS_TF_CRLF);
	//else TaltOS_long_output_write(TALTOS_TA_CRLF);
}


//The input buffer's output function, can be called by the tasks to receive data from out of the system
unsigned int TaltOS_long_input_read(unsigned long* ret)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	if (env->input_buf_start == env->input_buf_end){return TALTOS_FALSE;}
	(*ret) = *(((unsigned long*)TALTOS_INPUT_BUFFER_PTR) + env->input_buf_start);

	if (env->input_buf_start < TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long)) {env->input_buf_start++;}
	else {env->input_buf_start = 0;}
	(env->input_buf_act_size)--;
	return TALTOS_TRUE;
}

//The output buffer's input function, can be called by the tasks to send data out of system
void TaltOS_long_output_write(unsigned long inlong)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	unsigned long* address = (unsigned long*)(TALTOS_OUTPUT_BUFFER_PTR) + env->output_buf_end;
	(*address) = inlong;

	if (env->output_buf_end < TALTOS_OUTPUT_BUFFER_SIZE / sizeof(unsigned long))
	{
		if (env->output_buf_start != env->output_buf_end + 1){env->output_buf_end++;}
	}
	else
	{
		if (env->output_buf_start != 0){env->output_buf_end = 0;}
	}

	(env->output_buf_act_size)++;

	//LPC2106 specific code
	lpc2106_force_uart_irq();
	//LPC2106 specific code end
}

//The output buffer's output function, can be called by the microcontroller
unsigned int TaltOS_long_output_read(unsigned long* ret)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	if (env->output_buf_start == env->output_buf_end){return TALTOS_FALSE;}
	(*ret) = *((unsigned long*)TALTOS_OUTPUT_BUFFER_PTR + env->output_buf_start);

	if (env->output_buf_start < TALTOS_OUTPUT_BUFFER_SIZE / sizeof(unsigned long)) {env->output_buf_start++;}
	else {env->output_buf_start = 0;}
	(env->output_buf_act_size)--;
	return TALTOS_TRUE;
}


//Get the component's address by it's ID (Absolute address)
unsigned long* TaltOS_get_component_address(unsigned long comp_id1, unsigned long comp_id2)
{
	unsigned int isfound = TALTOS_FALSE;
	TaltOS_reg_entry *comp_reg;
	unsigned int i;
	for(i = 0; (i < TALTOS_COMPONENT_TABLE_SIZE/sizeof(TaltOS_reg_entry)) && (!isfound); i++)
	{
		comp_reg = (TaltOS_reg_entry*)TALTOS_COMPONENT_TABLE_PTR + i;
		if ((comp_reg->ID1 == comp_id1) && (comp_reg->ID2 == comp_id2))
		{
			isfound = TALTOS_TRUE;
		}
	}
	if (isfound) {return comp_reg->address;}
	else {return 0;}
}

//Get the component's n. function's pointer (Absolute address)
unsigned long TaltOS_get_component_function(unsigned long* address, unsigned long index)
{
	return (unsigned long)(address + (*(address + index)) / sizeof(unsigned long));
}

//Starts an own task from an own function
void TaltOS_add_own_task(unsigned long index, unsigned long argument)
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	unsigned long function_address;
	function_address = TaltOS_get_component_function(env->TaltOS_starting_address, index);
	if (!TaltOS_add_task(function_address, TALTOS_OWN_TASK_PRIORITY, argument)) TaltOS_long_output_write(TALTOS_TF_CRLF);
	//else TaltOS_long_output_write(TALTOS_TA_CRLF);
}


//Check if a component has arrived into the input buffer and handles component receiving
void TaltOS_check_order()
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	unsigned long order;
	if(env->input_buf_act_size >= 1)
	{
		order = *((unsigned long*)(TALTOS_INPUT_BUFFER_PTR) + env->input_buf_end - 1);
		
		switch (env->component_receiving_state)
		{
			case (COMPONENT_NOT_RECEIVING)	:
			{
				switch (order)
				{
					case (TALTOS_RUN_INIT)	:
					{
						TaltOS_add_own_task(TALTOS_RUN_AND_LIST_COMPONENTS_TASK, TALTOS_RUN_AND_LIST_SHOULDRUN);
						TaltOS_end_order();
						break;
					}
					case (TALTOS_COMP_BEGIN): 
					{
						env->component_receiving_state = WAITING_FOR_COMPONENT_SIZE;
						TaltOS_end_order();
						break;
					}
					case (TALTOS_COMP_END)	: 
					{
						TaltOS_long_output_write(TALTOS_CF_CRLF);
						TaltOS_end_order();
						break;
					}
					case (TALTOS_REM_COMP)	: 
					{
						TaltOS_end_order();
						env->removing_component_ID2 = TaltOS_end_order();
						env->removing_component_ID1 = TaltOS_end_order();
						TaltOS_add_own_task(TALTOS_REMOVE_COMPONENT, TALTOS_TASK_NULL_ARGUMENT);
						break;
					}
					case (TALTOS_COMP_LIST)	: 
					{
						TaltOS_add_own_task(TALTOS_RUN_AND_LIST_COMPONENTS_TASK, TALTOS_RUN_AND_LIST_SHOULDLIST);
						TaltOS_end_order();
						break;
					}
				}
				break;
			}
			case (WAITING_FOR_COMPONENT_SIZE)	:
			{
				env->component_receiving_size = order;
				TaltOS_end_order();
				env->component_receiving_state = COMPONENT_RECEIVING;
				break;
			}
			case (COMPONENT_RECEIVING)	:
			{
				if (order == TALTOS_COMP_END)
				{
					if (((env->component_receiving_size / sizeof(unsigned long))) == env->flash_op_buffer_ptr)
					{
						env->component_receiving_state = COMPONENT_NOT_RECEIVING;
						TaltOS_add_own_task(TALTOS_COPY_COMPONENT, TALTOS_TASK_NULL_ARGUMENT);
					}
					else
					{
						env->component_receiving_state = COMPONENT_NOT_RECEIVING;
						TaltOS_long_output_write(TALTOS_CF_CRLF);
						TaltOS_clear_flash_op_buffer();
					}
				}
				else
				{
					if ((env->component_receiving_size / sizeof(unsigned long)) < env->flash_op_buffer_ptr)
					{
						env->component_receiving_state = COMPONENT_NOT_RECEIVING;
						TaltOS_long_output_write(TALTOS_CF_CRLF);
						TaltOS_clear_flash_op_buffer();
					}
					else
					{
						*(((unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR) + env->flash_op_buffer_ptr) = order;
						env->flash_op_buffer_ptr++;
						if ((env->flash_op_buffer_ptr * sizeof(unsigned long)) == TALTOS_FLASH_OP_BUFFER_SIZE)
						{
							env->component_receiving_state = COMPONENT_NOT_RECEIVING;
							TaltOS_long_output_write(TALTOS_CF_CRLF);
							TaltOS_clear_flash_op_buffer();
						}
					}
				}
				TaltOS_end_order();
				break;
			}
		}
	}
}

//Removes the last long value from the input buffer (this was an order)
unsigned long TaltOS_end_order()
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	if (env->input_buf_end == 0) {env->input_buf_end = TALTOS_INPUT_BUFFER_SIZE / sizeof(unsigned long);}
	else {env->input_buf_end--;}
	env->input_buf_act_size--;
	return *((unsigned long*)(TALTOS_INPUT_BUFFER_PTR) + env->input_buf_end);
}


//Checks if the component's ids are garbage (wrong value => error in erase operation)
unsigned int TaltOS_is_garbage_reg(TaltOS_reg_entry *test_reg)
{
	unsigned char* test_array = (unsigned char*)test_reg;
	if (test_array[0] == TALTOS_COMPID_END && test_array[4] == TALTOS_COMPID_END) {return 0;}
	else return 1;
}

//Copys the component in the Flash operating memory to the Flash memory with first fit algorithm
void TaltOS_copy_component()
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	TaltOS_reg_entry component;
	unsigned int address_found = TALTOS_FALSE;
	unsigned int i = 0;
	unsigned int j = 0;
	unsigned long* empty_tester;
	unsigned long init_function_address;
	
	TaltOS_reg_entry *TaltOS_reg_ptr;
	
	//Get the ID's of the component
	component.ID1 = *(((unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR) + env->flash_op_buffer_ptr - 1);
	*(((unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR) + env->flash_op_buffer_ptr - 1) = TALTOS_EMPTY;
	component.ID2 = *(((unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR) + env->flash_op_buffer_ptr - 2);
	*(((unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR) + env->flash_op_buffer_ptr - 2) = TALTOS_EMPTY;
	
	//Get the size of the component
	component.size = ((env->flash_op_buffer_ptr - 2) / LPC2106_MINDATA) + 1;
	
	//Search for the first empty place of the TaltOS
	for(i = 0; (i < LPC2106_FLASH_MEMORY_SIZE / LPC2106_MINDATA) && (!address_found); i++)
	{
		address_found = TALTOS_TRUE;
		// Check if the full size is fit to this place
		for(j = 0; (j < component.size) && (address_found); j++)
		{
			empty_tester = (unsigned long*)((i + j) * LPC2106_MINDATA);
			if ((*empty_tester) != TALTOS_EMPTY)
			{
				address_found = TALTOS_FALSE;
			}
		}
	}
	
	if (address_found)
	{
		component.address = (unsigned long*)((i - 1) * LPC2106_MINDATA);

		//Copy the buffer by LPC2106_MINDATA size blocks
		for(j = 0; j < component.size; j++)
		{
			lpc2106_copy_to_flash_512(	(unsigned long)component.address + (j * LPC2106_MINDATA),
										(unsigned long*)(TALTOS_FLASH_OP_BUFFER_PTR + j * LPC2106_MINDATA),
										1);
		}

		TaltOS_clear_flash_op_buffer();
		
		//Save the full component table, and register the TaltOS if possible
		
		address_found = TALTOS_FALSE;
		for(i = 0; (i < (TALTOS_FLASH_OP_BUFFER_SIZE / sizeof(TaltOS_reg_entry))) && (!address_found); i++)
		{
			TaltOS_reg_ptr = ((TaltOS_reg_entry*)TALTOS_COMPONENT_TABLE_PTR) + i;
			if ((TaltOS_reg_ptr->ID1 == component.ID1) && (TaltOS_reg_ptr->ID2 == component.ID2))
			{
				TaltOS_reg_ptr = (TaltOS_reg_entry*)(TALTOS_FLASH_OP_BUFFER_PTR) + i;
				TaltOS_reg_ptr->ID1= ~(component.ID1);
				TaltOS_reg_ptr->ID2= ~(component.ID2);
				// We don't erase the address and size, for later garbage collection
			}
			else if ((TaltOS_reg_ptr->ID1 == TALTOS_EMPTY) && (TaltOS_reg_ptr->ID2 == TALTOS_EMPTY))
			{
				TaltOS_reg_ptr = ((TaltOS_reg_entry*)TALTOS_FLASH_OP_BUFFER_PTR) + i;
				TaltOS_reg_ptr->ID1 = component.ID1;
				TaltOS_reg_ptr->ID2 = component.ID2;
				TaltOS_reg_ptr->address = component.address;
				TaltOS_reg_ptr->size = component.size;
				address_found = TALTOS_TRUE;
			}
		}

		//Copy the full component table back to the flash
		lpc2106_copy_to_flash_512(TALTOS_COMPONENT_TABLE_PTR,(unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR,16);

		TaltOS_clear_flash_op_buffer();
		
		init_function_address = TaltOS_get_component_function(component.address, COMP_INIT);
		
		if (!TaltOS_add_task(init_function_address, TALTOS_INIT_TASK_PRIORITY, TALTOS_TASK_NULL_ARGUMENT)) TaltOS_long_output_write(TALTOS_TF_CRLF);
		//else TaltOS_long_output_write(TALTOS_TA_CRLF);
		
		TaltOS_long_output_write(TALTOS_CS_CRLF);
	}
	else
	{
		TaltOS_clear_flash_op_buffer();
		TaltOS_long_output_write(TALTOS_CF_CRLF);
	}
}

//Removes the component given by IDs
void TaltOS_remove_component()
{
	TaltOS_env* env = (TaltOS_env*)TALTOS_ENV_PTR;
	TaltOS_reg_entry* TaltOS_reg_ptr;
	unsigned int i;
	
	//Check the component table, if the specifid IDs are found, write the complement if them to the buffer
	for(i = 0; i < (TALTOS_FLASH_OP_BUFFER_SIZE / sizeof(TaltOS_reg_entry)); i++)
	{
		TaltOS_reg_ptr = (TaltOS_reg_entry*)(TALTOS_COMPONENT_TABLE_PTR) + i;
		if ((TaltOS_reg_ptr->ID1 == env->removing_component_ID1) && (TaltOS_reg_ptr->ID2 == env->removing_component_ID2))
		{
			TaltOS_reg_ptr = (TaltOS_reg_entry*)(TALTOS_FLASH_OP_BUFFER_PTR) + i;
			TaltOS_reg_ptr->ID1= ~(env->removing_component_ID1);
			TaltOS_reg_ptr->ID2= ~(env->removing_component_ID2);
			// We don't erase the address and size, for later garbage collection
		}
	}
	lpc2106_copy_to_flash_512(TALTOS_COMPONENT_TABLE_PTR,(unsigned long*)TALTOS_FLASH_OP_BUFFER_PTR,16);

	TaltOS_clear_flash_op_buffer();

	env->removing_component_ID1 = 0;
	env->removing_component_ID2 = 0;
	
	TaltOS_long_output_write(TALTOS_CR_CRLF);
}

