/*
	TaltOS Definitions
	on the Philips Chip LPC2106

	By Peter Eordegh
*/



//Component table pointer
#define TALTOS_COMPONENT_TABLE_PTR (0x0001C000)	//Sector 14 (sector 15 is the remapped bootloader)
#define TALTOS_COMPONENT_TABLE_SIZE (0x2000)	//This is a full sector size

//LPC2106 sepcific flash memory parameters
#define LPC2106_FLASH_MEMORY_SIZE (0x20000)	//All the flash memory size
#define LPC2106_SECTOR_SIZE (0x2000)		//Sector size 8192 bytes
#define LPC2106_MINDATA (0x200)				//1/16 part of the sector can be written 512 bytes
#define LPC2106_RAM_TOP_PTR (0x4000FFE0)

//Flash operation buffer
#define TALTOS_FLASH_OP_BUFFER_SIZE (0x2000)	//This is a full sector size
#define TALTOS_FLASH_OP_BUFFER_PTR (LPC2106_RAM_TOP_PTR-TALTOS_FLASH_OP_BUFFER_SIZE)

//Input data buffer
#define TALTOS_INPUT_BUFFER_SIZE (0x800)		//This is a half of a sector size
#define TALTOS_INPUT_BUFFER_PTR (TALTOS_FLASH_OP_BUFFER_PTR-TALTOS_INPUT_BUFFER_SIZE)

//Output data buffer
#define TALTOS_OUTPUT_BUFFER_SIZE (0x800)		//This is a half of a sector size
#define TALTOS_OUTPUT_BUFFER_PTR (TALTOS_INPUT_BUFFER_PTR-TALTOS_OUTPUT_BUFFER_SIZE)

#define TALTOS_ENV_PTR (TALTOS_OUTPUT_BUFFER_PTR-sizeof(TaltOS_env))

#define TALTOS_TASK_LIST_SIZE (10)
#define TALTOS_TASK_LIST_PTR (TALTOS_ENV_PTR-(TALTOS_TASK_LIST_SIZE*sizeof(TaltOS_task)))


//TaltOS IDs
#define TALTOS_ID1 (0x0000014E)
#define TALTOS_ID2 (0x0000014E)

//TaltOS Functions (with some lpc2106 specific funtions for the vectored interrupt setup)
#define TALTOS_INIT						(0x00000000)
#define LPC2106_UART_HANDLER			(0x00000001)
#define LPC2106_TIMER_HANDLER			(0x00000002)
#define LPC2106_DEFAULT_HANDLER			(0x00000003)
#define TALTOS_CHECK_ORDER				(0x00000004)
#define TALTOS_RUN_AND_LIST_COMPONENTS_TASK	(0x00000005)
#define TALTOS_COPY_COMPONENT			(0x00000006)
#define TALTOS_REMOVE_COMPONENT			(0x00000007)
#define TALTOS_ADD_TASK					(0x00000008)
#define TALTOS_GET_COMPONENT_ADDRESS	(0x00000009)
#define TALTOS_GET_COMPONENT_FUNCTION	(0x0000000A)
#define TALTOS_LONG_OUTPUT_WRITE		(0x0000000B)

#define COMP_INIT						(0x00000000)

#define TALTOS_ERASED (0x00000000)
#define TALTOS_EMPTY (0xFFFFFFFF)
#define TALTOS_COMPID_END (0x4E)

//The characters to identify if a component has arrived

//Command Strings (unsigned longs)
#define TALTOS_COMP_BEGIN	(0x2A62632A)	// *cb* : Component Begin
#define TALTOS_COMP_END		(0x2A65632A)	// *ce* : Component End
#define TALTOS_COMP_LIST	(0x2A6C632A)	// *cl* : Component List
#define TALTOS_REM_COMP		(0x2A63722A)	// *rc* : Remove Component
#define TALTOS_RUN_INIT		(0x2A69722A)	// *ri* : Run (component) Init

//TaltOS replies
#define TALTOS_CRLF_OK		(0x0D0A4B4F)	// CR and LF than "OK" for : Command accepted
#define TALTOS_OK_CRLF		(0x4B4F0D0A)	// "OK" than CR and LF for : Command accepted
#define TALTOS_CRLFCRLF		(0x0D0A0D0A)	// CR and LF than again  : Just spacing 
#define TALTOS_00000000		(0x30303030)	// The allzero characters to make the ID's readable
#define TALTOS_TA_CRLF		(0x41540D0A)	// "TA" than CR and LF for : Task added to task list
#define TALTOS_TF_CRLF		(0x46540D0A)	// "TF" than CR and LF for : Tasklist is full
#define TALTOS_CS_CRLF		(0x53430D0A)	// "CS" than CR and LF for : Component upload was succesful
#define TALTOS_CF_CRLF		(0x46430D0A)	// "CF" than CR and LF for : Component upload failed
#define TALTOS_CR_CRLF		(0x52430D0A)	// "CR" than CR and LF for : Component removed


//Booleans
#define TALTOS_TRUE		(1)
#define TALTOS_FALSE	(0)

//TaltOS task states
#define TALTOS_TASK_OFF			0
#define TALTOS_TASK_READYTORUN	1
#define TALTOS_TASK_RUNNING		2
#define TALTOS_TASK_WAITING		3
#define TALTOS_TASK_SLEEPING	4

//Component arriving states

#define COMPONENT_NOT_RECEIVING		0
#define WAITING_FOR_COMPONENT_SIZE	1
#define COMPONENT_RECEIVING			2

//TaltOS task struct handling constants
#define TALTOS_TASK_NOADDRESS		0
#define TALTOS_TASK_NULL_PTR		0
#define TALTOS_TASK_NULL_PRIORITY	0
#define TALTOS_OWN_TASK_PRIORITY	64
#define TALTOS_INIT_TASK_PRIORITY	128
#define TALTOS_TASK_NULL_ARGUMENT	0

//TaltOS_run_and_list_components controllers
#define TALTOS_RUN_AND_LIST_SHOULDLIST	(0x00000001)
#define TALTOS_RUN_AND_LIST_SHOULDRUN	(0x00000002)

//Debug Macro

#define DEBUG(X) REG8(UART0_THR) = X;

typedef struct
{
	unsigned long address;
	unsigned short priority;
	unsigned short state;
	unsigned long argument;
} TaltOS_task;

typedef struct
{
	unsigned long TaltOS_get_component_address_ptr;
	unsigned long TaltOS_get_component_function_ptr;
	unsigned long argument;
} TaltOS_task_parameters;

typedef struct
{
	unsigned long ID1;
	unsigned long ID2;
	unsigned long* address;
	unsigned long size;
} TaltOS_reg_entry;

typedef struct
{
	unsigned short input_buf_start;
	unsigned short input_buf_end;
	unsigned short input_buf_act_size;
	unsigned short output_buf_start;
	unsigned short output_buf_end;
	unsigned short output_buf_act_size;
	unsigned long  flash_op_buffer_ptr;
	TaltOS_task* actual_task;
	unsigned long* TaltOS_starting_address;
	unsigned short component_receiving_state;
	unsigned short component_receiving_size;
	unsigned long  removing_component_ID1;
	unsigned long  removing_component_ID2;
} TaltOS_env;
